{% extends "!layout.html" %}

{% block relbar1 %}
  <a href="index.html">python-systemd </a>·
  <a href="http://www.freedesktop.org/software/systemd/man/index.html">man pages </a>·
  <a href="http://www.freedesktop.org/software/systemd/man/systemd.directives.html">directives </a>

  <span style="float:right">python-systemd {{release}}</span>
  <hr />
{% endblock %}

{# remove the lower relbar #}
{% block relbar2 %} {% endblock %}

{# remove the footer #}
{% block footer %} {% endblock %}
